from .checksums import generate_checksums
from .fix_file_exts import fix_extensions
from .identify_files import identify
from .path_utils import explore_dir

__all__ = ["generate_checksums", "fix_extensions", "identify", "explore_dir"]



# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

from acamodels import ACABase
from acamodels import ArchiveFile
from digiarch.database import FileDB
from pydantic import Field
from pydantic import root_validator

# -----------------------------------------------------------------------------
# Model
# -----------------------------------------------------------------------------


class FileData(ACABase):
    main_dir: Path
    data_dir: Path = Field(None)
    db: FileDB = Field(None)
    files: List[ArchiveFile]

    class Config:
        arbitrary_types_allowed = True

    @root_validator
    def create_dir(cls, fields: Dict[Any, Any]) -> Dict[Any, Any]:
        main_dir = fields.get("main_dir")
        data_dir = fields.get("data_dir")
        db = fields.get("db")
        if data_dir is None and main_dir:
            data_dir = main_dir / "_digiarch"
            data_dir.mkdir(exist_ok=True)
            fields["data_dir"] = data_dir
        if db is None and data_dir:
            db_path: Path = fields["data_dir"] / "files.db"
            fields["db"] = FileDB(f"sqlite:///{db_path}")
        return fields



"""Module level docstring.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import json
from pathlib import Path
from typing import List

from acamodels import ArchiveFile
from tqdm import tqdm

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def fix_extensions(files: List[ArchiveFile]) -> List[ArchiveFile]:
    map_path = Path(__file__).parents[1] / "_data" / "ext_map.json"
    ext_map = json.load(map_path.open(encoding="utf-8"))
    to_fix = [
        file
        for file in files
        if "Extension mismatch" in (file.warning or "")
        and file.puid in ext_map
    ]
    for file in tqdm(
        to_fix, desc="Fixing file extensions", disable=not to_fix
    ):
        new_name = file.path.with_name(f"{file.name()}.{ext_map[file.puid]}")
        file.path.rename(new_name)
        file.path = new_name

    return to_fix



"""Identify files using
`siegfried <https://github.com/richardlehane/siegfried>`_

"""
# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import json
import re
import subprocess
from functools import partial
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

from acamodels import ArchiveFile
from acamodels import Identification
from digiarch.core.utils import natsort_path
from digiarch.exceptions import IdentificationError

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def custom_id(path: Path, file_id: Identification) -> Identification:
    sig_lwp = re.compile(
        r"(?i)^576F726450726F0DFB000000000000"
        "000005985C8172030040CCC1BFFFBDF970"
    )
    sig_gif_bof = re.compile(r"(?i)^474946383961")
    sig_gif_eof = re.compile(r"(?i)3B")
    sig_123 = re.compile(r"(?i)^00001A000(3|4|5)10040000000000")
    sig_mmap = re.compile(r"(?i)4D696E644d616E61676572")
    sig_word_markup = re.compile(
        r"(?i)(50|70)726F67(49|69)64[0-9A-F]{2,20}576f72642e446f63756d656e74"
    )
    sig_excel_markup = re.compile(
        r"(?i)(50|70)726F67(49|69)64[0-9A-F]{2,18}457863656C2E5368656574"
    )
    with path.open("rb") as file_bytes:
        # BOF
        bof = file_bytes.read(1024).hex()
        # Navigate to EOF
        try:
            file_bytes.seek(-1024, 2)
        except OSError:
            # File too small :)
            file_bytes.seek(-file_bytes.tell(), 2)
        eof = file_bytes.read(1024).hex()

        if sig_lwp.match(bof):
            file_id.puid = "x-fmt/340"
            file_id.signature = "Lotus WordPro Document"
            if path.suffix.lower() != ".lwp":
                file_id.warning = "Extension mismatch"
            else:
                file_id.warning = None
        elif sig_123.match(bof):
            file_id.puid = "aca-fmt/1"
            file_id.signature = "Lotus 1-2-3 Spreadsheet"
            if path.suffix.lower() != ".123":
                file_id.warning = "Extension mismatch"
            else:
                file_id.warning = None
        elif sig_word_markup.search(bof):
            file_id.puid = "aca-fmt/2"
            file_id.signature = "Microsoft Word Markup"
            if path.suffix.lower() != ".doc":
                file_id.warning = "Extension mismatch"
            else:
                file_id.warning = None
        elif sig_excel_markup.search(bof):
            file_id.puid = "aca-fmt/3"
            file_id.signature = "Microsoft Excel Markup"
            if path.suffix.lower() != ".xls":
                file_id.warning = "Extension mismatch"
            else:
                file_id.warning = None
        elif sig_mmap.search(bof) or sig_mmap.search(eof):
            file_id.puid = "aca-fmt/4"
            file_id.signature = "MindManager Mind Map"
            if path.suffix.lower() != ".mmap":
                file_id.warning = "Extension mismatch"
            else:
                file_id.warning = None
        elif sig_gif_bof.match(bof) and sig_gif_eof.search(eof):
            file_id.puid = "fmt/4"
            file_id.signature = "Graphics Interchange Format"
            if path.suffix.lower() != ".gif":
                file_id.warning = "Extension mismatch"
            else:
                file_id.warning = None
    return file_id


def sf_id(path: Path) -> Dict[Path, Identification]:
    """Identify files using
    `siegfried <https://github.com/richardlehane/siegfried>`_ and update
    FileInfo with obtained PUID, signature name, and warning if applicable.

    Parameters
    ----------
    path : pathlib.Path
        Path in which to identify files.

    Returns
    -------
    Dict[Path, Identification]
        Dictionary containing file path and associated identification
        information obtained from siegfried's stdout.

    Raises
    ------
    IdentificationError
        If running siegfried or loading of the resulting JSON output fails,
        an IdentificationError is thrown.
    """

    id_dict: Dict[Path, Identification] = {}

    try:
        sf_proc = subprocess.run(
            ["sf", "-json", "-multi", "1024", str(path)],
            check=True,
            capture_output=True,
        )
    except Exception as error:
        raise IdentificationError(error)

    try:
        id_result = json.loads(sf_proc.stdout)
    except Exception as error:
        raise IdentificationError(error)

    for file_result in id_result.get("files", []):
        match: Dict[str, Any] = {}
        for id_match in file_result.get("matches"):
            if id_match.get("ns") == "pronom":
                match = id_match
        if match:
            file_identification: Identification
            file_path: Path = Path(file_result["filename"])

            if match.get("id", "").lower() == "unknown":
                puid = None
            else:
                puid = match.get("id")

            signature = match.get("format")
            warning = match.get("warning", "").capitalize()
            file_identification = Identification(
                puid=puid, signature=signature or None, warning=warning or None
            )
            if puid is None:
                file_identification = custom_id(file_path, file_identification)
            if (
                puid in ["fmt/96", "fmt/101", "fmt/583", "x-fmt/263"]
                and "Extension mismatch" in warning
            ):
                file_identification = custom_id(file_path, file_identification)
            id_dict.update({file_path: file_identification})

    return id_dict


def update_file_info(
    file_info: ArchiveFile, id_info: Dict[Path, Identification]
) -> ArchiveFile:
    no_id: Identification = Identification(
        puid=None,
        signature=None,
        warning="No identification information obtained.",
    )
    new_id: Identification = id_info.get(file_info.path) or no_id
    if file_info.path.stat().st_size == 0:
        new_id = Identification(
            puid=None,
            signature=None,
            warning="File is empty.",
        )
    file_info = file_info.copy(update=new_id.dict())
    return file_info


def identify(files: List[ArchiveFile], path: Path) -> List[ArchiveFile]:
    """Identify all files in a list, and return the updated list.

    Parameters
    ----------
    files : List[FileInfo]
        Files to identify.

    Returns
    -------
    List[FileInfo]
        Input files with updated Identification information.

    """

    id_info: Dict[Path, Identification] = sf_id(path)
    _update = partial(update_file_info, id_info=id_info)
    updated_files: List[ArchiveFile] = list(map(_update, files))

    return natsort_path(updated_files)



# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

from typing import List

from acamodels import ArchiveFile
from natsort import natsorted

# -----------------------------------------------------------------------------
# Utility functions
# -----------------------------------------------------------------------------


def size_fmt(size: float) -> str:
    """Formats a file size in binary multiples to a human readable string.

    Parameters
    ----------
    size : float
        The file size in bytes.

    Returns
    -------
    str
        Human readable string representing size in binary multiples.
    """
    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if size < 1024.0:
            break
        size /= 1024.0
    return f"{size:.1f} {unit}"


def natsort_path(file_list: List[ArchiveFile]) -> List[ArchiveFile]:
    """Naturally sort a list of FileInfo objects by their paths.

    Parameters
    ----------
    file_list : List[FileInfo]
        The list of FileInfo objects to be sorted.

    Returns
    -------
    List[ArchiveFile]
        The list of FileInfo objects naturally sorted by their path.
    """

    sorted_file_list: List[ArchiveFile] = natsorted(
        file_list, key=lambda archive_file: str(archive_file.path)
    )

    return sorted_file_list



"""Utilities for handling files, paths, etc.

"""
# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import List

from acamodels import ArchiveFile
from digiarch.core.utils import natsort_path
from digiarch.core.utils import size_fmt
from digiarch.exceptions import FileCollectionError
from digiarch.models import FileData
from digiarch.models import Metadata
from tqdm import tqdm

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


async def explore_dir(file_data: FileData) -> List[str]:
    """Finds files and empty directories in the given path,
    and writes them to a file database.

    Parameters
    ----------
    file_data: models.FileData
        File data model including main directory, data directory,
        and file database.

    Returns
    ----------
    List[str]
        List of warnings encountered while parsing directories.

    """
    # Type declarations
    dir_info: List[ArchiveFile] = []
    empty_subs: List[Path] = []
    multiple_files: List[Path] = []
    warnings: List[str] = []
    total_size: int = 0
    file_count: int = 0
    main_dir: Path = file_data.main_dir
    data_dir: Path = file_data.data_dir
    metadata = Metadata(
        last_run=datetime.now(), processed_dir=file_data.main_dir
    )

    if not [
        child for child in main_dir.iterdir() if child.name != data_dir.name
    ]:
        # Path is empty, remove main directory and raise
        shutil.rmtree(data_dir)
        raise FileCollectionError(f"{main_dir} is empty! No files collected.")

    # Traverse given path, collect results.
    # tqdm is used to show progress of os.walk
    for root, dirs, files in tqdm(
        os.walk(main_dir, topdown=True), unit=" folders", desc="Processed"
    ):
        if data_dir.name in dirs:
            # Don't walk the _digiarch directory
            dirs.remove(data_dir.name)
        if not dirs and not files:
            # We found an empty subdirectory.
            empty_subs.append(Path(root))
        if len(files) > 1:
            multiple_files.append(Path(root))
        for file in files:
            cur_path = Path(root, file)
            dir_info.append(ArchiveFile(path=cur_path))
            total_size += cur_path.stat().st_size
            file_count += 1

    dir_info = natsort_path(dir_info)

    # Update metadata
    metadata.file_count = file_count
    metadata.total_size = size_fmt(total_size)

    # Update aux tables
    if empty_subs:
        await file_data.db.set_empty_subs(empty_subs)
        warnings.append("Warning! Empty subdirectories detected!")
    if multiple_files:
        await file_data.db.set_multi_files(multiple_files)
        warnings.append("Warning! Some directories have multiple files!")

    # Update db
    await file_data.db.set_metadata(metadata)
    await file_data.db.set_files(dir_info)

    return warnings



"""This implements the Command Line Interface which enables the user to
use the functionality implemented in the :mod:`~digiarch` submodules.
The CLI implements several commands with suboptions.

"""
# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import asyncio
from functools import wraps
from pathlib import Path
from typing import Any
from typing import Callable
from typing import List

import click
from click.core import Context
from digiarch import __version__
from digiarch import core
from digiarch.exceptions import FileCollectionError
from digiarch.exceptions import FileParseError
from digiarch.exceptions import IdentificationError
from digiarch.models import FileData

# -----------------------------------------------------------------------------
# Auxiliary functions
# -----------------------------------------------------------------------------


def coro(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


# -----------------------------------------------------------------------------
# CLI
# -----------------------------------------------------------------------------


@click.group(invoke_without_command=True, chain=True)
@click.argument(
    "path", type=click.Path(exists=True, file_okay=False, resolve_path=True)
)
@click.option("--reindex", is_flag=True, help="Reindex the current directory.")
@click.version_option(version=__version__)
@click.pass_context
@coro
async def cli(ctx: Context, path: str, reindex: bool) -> None:
    """Used for indexing, reporting on, and identifying files
    found in PATH.
    """

    # Initialise
    in_path: Path = Path(path)
    file_data: FileData = FileData(main_dir=in_path, files=[])
    empty: bool = await file_data.db.is_empty()
    warnings: List[str] = []

    # Collect file info and update file_data
    if reindex or empty:
        click.secho("Collecting file information...", bold=True)
        try:
            warnings = await core.explore_dir(file_data)
        except FileCollectionError as error:
            raise click.ClickException(str(error))

    else:
        click.echo("Processing data from ", nl=False)
        click.secho(f"{file_data.db.url}", bold=True)

    for warning in warnings:
        click.secho(warning, bold=True, fg="red")

    try:
        file_data.files = await file_data.db.get_files()
    except FileParseError as error:
        raise click.ClickException(str(error))
    else:
        ctx.obj = file_data


@cli.command()
@click.pass_obj
def process(file_data: FileData) -> None:
    """Generate checksums and identify files."""
    _files = file_data.files
    _files = core.generate_checksums(_files)
    click.secho("Identifying files... ", nl=False)
    try:
        _files = core.identify(_files, file_data.main_dir)
    except IdentificationError as error:
        raise click.ClickException(str(error))
    else:
        click.secho(f"Successfully identified {len(_files)} files.")
        file_data.files = _files


@cli.command()
@click.pass_context
@coro
async def fix(ctx: Context) -> None:
    """Fix file extensions - files should be identified first."""
    file_data = ctx.obj
    fixed = core.fix_extensions(file_data.files)
    if fixed:
        click.secho("Rebuilding file information...", bold=True)
        new_files = core.identify(fixed, file_data.main_dir)
        await file_data.db.update_files(new_files)
        file_data.files = await file_data.db.get_files()
        ctx.obj = file_data
    else:
        click.secho("Info: No file extensions to fix.", bold=True, fg="yellow")


@cli.resultcallback()
@coro
async def done(result: Any, **kwargs: Any) -> None:
    ctx = click.get_current_context()
    file_data: FileData = ctx.obj
    await file_data.db.set_files(file_data.files)
    click.secho("Done!", bold=True, fg="green")



"""File database backend"""
# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import re
from pathlib import Path
from typing import Any
from typing import List

import sqlalchemy as sql
from acamodels import ArchiveFile
from databases import Database
from digiarch.exceptions import FileParseError
from digiarch.models.metadata import Metadata
from pydantic import parse_obj_as
from pydantic import ValidationError
from sqlalchemy.exc import OperationalError
from sqlalchemy_utils import create_view

# -----------------------------------------------------------------------------
# Database class
# -----------------------------------------------------------------------------


class FileDB(Database):
    """File database"""

    sql_meta = sql.MetaData()

    metadata = sql.Table(
        "Metadata",
        sql_meta,
        sql.Column("last_run", sql.DateTime, nullable=False),
        sql.Column("processed_dir", sql.String, nullable=False),
        sql.Column("file_count", sql.Integer),
        sql.Column("total_size", sql.Integer),
    )

    files = sql.Table(
        "Files",
        sql_meta,
        sql.Column("id", sql.Integer, primary_key=True, autoincrement=True),
        sql.Column("uuid", sql.String, nullable=False),
        sql.Column("path", sql.String, nullable=False),
        sql.Column("checksum", sql.String),
        sql.Column("aars_path", sql.String, nullable=False),
        sql.Column("puid", sql.String),
        sql.Column("signature", sql.String),
        sql.Column("warning", sql.String),
    )

    multiple_files = sql.Table(
        "_MultipleFiles",
        sql_meta,
        sql.Column("path", sql.String, nullable=False),
    )

    empty_directories = sql.Table(
        "_EmptyDirectories",
        sql_meta,
        sql.Column("path", sql.String, nullable=False),
    )

    id_warnings = files.select().where(files.c.warning.isnot(None))
    puid_none = sql.case(
        [(files.c.puid.is_(None), "None")],
        else_=files.c.puid,
    )
    sig_count = (
        sql.select(
            [
                files.c.puid,
                files.c.signature,
                sql.func.count(puid_none).label("count"),
            ]
        )
        .group_by("puid")
        .order_by(sql.desc("count"))
    )
    create_view("_IdentificationWarnings", id_warnings, sql_meta)
    create_view("_SignatureCount", sig_count, sql_meta)

    def __init__(self, url: str) -> None:
        super().__init__(url)
        engine = sql.create_engine(
            url, connect_args={"check_same_thread": False}
        )
        try:
            self.sql_meta.create_all(engine)
        except OperationalError as error:
            warn_re = re.compile(
                r"(?i)(IdentificationWarnings|SignatureCount)"
            )
            if warn_re.search(str(error)):
                pass
            else:
                raise

    async def is_empty(self) -> bool:
        query = self.files.select()
        result = await self.fetch_one(query)
        if result is None:
            return True
        else:
            return False

    async def delsert(self, table: sql.Table, values: Any) -> None:
        delete = table.delete()
        insert = table.insert()
        async with self.transaction():
            await self.execute(query=delete)
            if isinstance(values, list):
                await self.execute_many(query=insert, values=values)
            else:
                await self.execute(query=insert, values=values)

    async def set_metadata(self, metadata: Metadata) -> None:
        meta = {
            "processed_dir": str(metadata.processed_dir),
            **metadata.copy(exclude={"processed_dir"}).dict(),
        }
        await self.delsert(self.metadata, values=meta)

    async def set_files(self, files: List[ArchiveFile]) -> None:
        encoded_files = [file.encode() for file in files]
        await self.delsert(self.files, values=encoded_files)

    async def update_files(self, new_files: List[ArchiveFile]) -> None:
        encoded_files = [file.encode() for file in new_files]
        async with self.transaction():
            for file in encoded_files:
                update = (
                    self.files.update()
                    .where(self.files.c.uuid == file["uuid"])
                    .values(file)
                )
                await self.execute(update)

    async def get_files(self) -> List[ArchiveFile]:
        query = self.files.select()
        rows = await self.fetch_all(query)
        try:
            files = parse_obj_as(List[ArchiveFile], rows)
        except ValidationError:
            raise FileParseError(
                "Failed to parse files as ArchiveFiles. Please reindex."
            )
        return files

    async def set_multi_files(self, multi_files: List[Path]) -> None:
        await self.delsert(
            self.multiple_files,
            values=[{"path": str(path)} for path in multi_files],
        )

    async def set_empty_subs(self, empty_subs: List[Path]) -> None:
        await self.delsert(
            self.empty_directories,
            values=[{"path": str(path)} for path in empty_subs],
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['digiarch', 'digiarch.core', 'digiarch.database', 'digiarch.models']

package_data = \
{'': ['*'], 'digiarch': ['_data/*']}

install_requires = \
['acamodels>=0.5.3,<0.6.0',
 'click>=7.0,<8.0',
 'dacite>=1.0,<2.0',
 'databases[sqlite]>=0.3.2,<0.5.0',
 'natsort>=7.0.1,<8.0.0',
 'pydantic>=1.6.1,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests>=2.23.0,<3.0.0',
 'sqlalchemy>=1.3.19,<2.0.0',
 'sqlalchemy_utils>=0.36.8,<0.37.0',
 'tqdm>=4.36,<5.0']

entry_points = \
{'console_scripts': ['digiarch = digiarch.cli:cli']}

setup_kwargs = {
    'name': 'digiarch',
    'version': '0.9.1',
    'description': 'Tools for the Digital Archive Project at Aarhus Stadsarkiv',
    'long_description': '[![Aarhus Stadsarkiv](https://raw.githubusercontent.com/aarhusstadsarkiv/py-template/master/img/logo.png)](https://stadsarkiv.aarhus.dk/)\n# Digital Archive [![Tests](https://github.com/aarhusstadsarkiv/digiarch/workflows/Tests/badge.svg?branch=master)](https://github.com/aarhusstadsarkiv/digiarch/actions?query=workflow%3ATests+branch%3Amaster) [![codecov](https://codecov.io/gh/aarhusstadsarkiv/digiarch/branch/master/graph/badge.svg)](https://codecov.io/gh/aarhusstadsarkiv/digiarch) [![PyPI](https://badge.fury.io/py/digiarch.svg)](https://badge.fury.io/py/digiarch)\nThis repository contains code pertaining to the Digital Archive Project at Aarhus Stadsarkiv.\n',
    'author': 'Nina Jensen',
    'author_email': 'jnik@aarhus.dk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://stadsarkiv.aarhus.dk/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Custom exceptions defined for use in digiarch modules.

"""

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------


class DigiarchError(Exception):
    """Base class for digiarch errors."""


class IdentificationError(DigiarchError):
    """Implements an error to raise if identification or related
    functionality fails."""


class FileCollectionError(DigiarchError):
    """Implements an error to raise if File discovery/collection or related
    functionality fails."""


class FileParseError(DigiarchError):
    """Implements an error to raise if file parsing fails"""
